import os
import requests
import warnings
from pathlib import Path
from pprint import pformat
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor


class Collections:
    def __init__(self,ummResponse):
        """
        Collections class that extracts the most relevant metadata properties from
        CMR collections and stores them as a dictionary for all collections in reponse
        """

        # tuple of keys to extract from the server response for collections
        keepKeys = (
            'native-id',
            'concept-id',
            'provider-id',
            'ShortName',
            'EntryTitle',
            'ProcessingLevel',
            'SpatialExtent',
            'ScienceKeywords',
            'TemporalExtents',
            'Abstract',
            'RelatedUrls',
            'revision-date',
            'Version',
        )

        topLevelKeys = list(ummResponse['items'][0].keys())

        collections = []
        for item in ummResponse['items']:
            md = item[topLevelKeys[0]]
            for k in topLevelKeys[1:]:
                md.update(item[k])
            metadataKeys = md.keys()
            collections.append(OrderedDict({k: md[k] for k in keepKeys if k in metadataKeys}))

        self.items = collections

        return

    def __repr__(self):
        strRepr = pformat(self.items, depth=4)
        return f'CMR Collection Query result:\n{strRepr}'


class Granules:
    def __init__(self, ummResponse):
        """
        Granules class that extracts the most relevant metadata properties from
        CMR granules and stores them as a dictionary for all granules in reponse
        """

        # tuple of keys to extract from the server response for granules
        keepKeys = (
            'concept-id',
            'RelatedUrls',
            'SpatialExtent',
            'TemporalExtent',
            'GranuleUR',
            'revision-date'
        )

        topLevelKeys = list(ummResponse['items'][0].keys())

        granules = []
        for item in ummResponse['items']:
            md = item[topLevelKeys[0]]
            for k in topLevelKeys[1:]:
                md.update(item[k])

            metadataKeys = md.keys()
            granules.append(OrderedDict({k: md[k] for k in keepKeys if k in metadataKeys}))

        self.items = granules

        return

    def __repr__(self):
        strRepr = pformat(self.items, depth=4)
        return f'CMR Granule Query result:\n{strRepr}'

    @property
    def length(self):
        return len(self.items)

    def fetch(self, credentials, directory='./', limit=None, maxWorkers=2):
        """
        Method to pull the granules from the CMR response to local directory
        uses asynchronous calls to speed up downloading multiple files

        Args:
            credentials (tuple|list): Earthdata username and password needed to download data

        Kwargs:
            directory (string): relative or absolute path to download files too,
                default = current working directory
            limit (int): maximum number of granules to download,
                default = None (get all granules)
            marWorkers (int): maximum threads to execute granule fetching asynchronously,
                default = 2

        Returns:
            None
        """

        if maxWorkers < 1:
            warnings.warn("the specified maxWorkers is less than the minimum (1) "
                "using at least 1 worker")
            maxWorkers = 1


        if limit is not None:
            maxIters = min(limit, self.length)
        else:
            maxIters = self.length

        outDir = Path(directory)

        itemsToGet = self.items[:maxIters]

        # send download requests asynchronously
        with ThreadPoolExecutor(maxWorkers) as executor:
            executor.map(lambda x: self._granule_request(
                x, credentials, outDir), itemsToGet)

        return

    def getLocalPaths(self,directory='./', limit=None):
        """
        Method to construct local paths from granules,
        used for getting a potential list of files that are downloaded from `Granules.fetch()`

        Args:
            None

        Kwargs:
            directory (string): relative or absolute path to download files too,
                default = current working directory
            limit (int): maximum number of granules to download,
                default = None (get all granules)

        Returns:
            paths (tuple): tuple of all local paths for granules
        """

        if limit is not None:
            maxIters = min(limit, len(self.items))
        else:
            maxIters = len(self.items)

        outDir = Path(directory)

        itemsToGet = self.items[:maxIters]

        paths = []
        for item in itemsToGet:
            source = self._parseDataSource(item)
            _, fileName = os.path.split(source['URL'])
            outPath = outDir / fileName
            paths.append(outPath)

        return tuple(paths)

    @staticmethod
    def _granule_request(item, credentials, outDir):
        """
        Helper method to request granule download, used for asynchronous calls

        Args:
            item (dict): Granule response
            credentials(list|tuple): Earthdata username and password needed to download data
            outDir (pathlib.Path): out directory to save data too

        Returns:
            None
        """

        source = Granules._parseDataSource(item)
        with requests.Session() as s:
            s.auth = credentials
            authGateway = s.request('get', source['URL'])
            r = s.request('get',authGateway.url, auth=credentials)

            _, fileName = os.path.split(source['URL'])
            outFile = outDir / fileName
            outFile.write_bytes(r.content)

        return

    @staticmethod
    def _parseDataSource(item):
        """
        Helper method to extract granule download URL from response dictionary

        Args:
            item (dict): Granule response

        Returns:
            source (dict): Dictionary with key of `URL` that contains the granule
                download url

        """
        urls = item['RelatedUrls']
        source = dict([('URL', url['URL'])
                       for url in urls if url['Type'] == 'GET DATA'])

        return source



import datetime
from collections import OrderedDict


def decode_date(string):
    """Decodes a date from a command line argument, returning datetime object".

    Args:
      string: See AssetSetCommand class comment for the allowable
        date formats.v

    Returns:
      long, datetime object
    Raises:
      ValueError: if string does not conform to a legal date format.
    """
    date_formats = ['%Y%m%d',
                    '%Y-%m-%d',
                    '%Y-%m-%dT%H:%M:%S',
                    '%Y-%m-%dT%H:%M:%S.%f']
    for date_format in date_formats:
        try:
            dt = datetime.datetime.strptime(string, date_format)
            return dt
        except ValueError:
            continue
    raise ValueError('Invalid format for date: "%s".' % string)



import os
import json
import time
import warnings
import datetime
import requests
import requests_cache
from pprint import pformat
from simplecmr import utils
from simplecmr.result import Collections, Granules


class Query:
    def __init__(self, conceptid=None, shortName=None, spatialExtent=None,
                 startTime=None, endTime=None, maxResults=10,cache=True):
        """

        """

        # setting basic parameters for a query
        self.BASE_URL = "https://cmr.earthdata.nasa.gov/search"
        self.FORMAT = 'umm_json_v1_4'

        if cache:
            expire_after = datetime.timedelta(hours=1)
            requests_cache.install_cache(backend='memory',expire_after=expire_after)

        payload = {}

        if conceptid is not None:
            payload['concept_id'] = conceptid

            if shortName is not None:
                warnings.warn("parameters conceptid and shortName are mutually "
                              "exclusive...using conceptid for search")
        elif shortName is not None:
            payload['short_name'] = shortName

        # validate that the maxResults keyword is within reasonable boundes
        if (maxResults > 2000) or (maxResults < 1):
            raise ValueError(
                "Specified maximum results must be greater than 1 and less than 2001")
        else:
            payload['page_size'] = maxResults

        # check the inputs for the datetime information
        if (startTime is not None) or (endTime is not None):
            temporalExtent = ','
            if startTime is not None:
                if type(startTime) is not datetime.datetime:
                    startTime = utils.decode_date(startTime)
                temporalExtent = startTime.strftime(
                    "%Y-%m-%dT%H:%M:%SZ") + temporalExtent
            if endTime is not None:
                if type(endTime) is not datetime.datetime:
                    endTime = utils.decode_date(endTime)
                temporalExtent = temporalExtent + \
                    endTime.strftime("%Y-%m-%dT%H:%M:%SZ")
            # add temporal information to the request parameters
            payload['temporal'] = temporalExtent

        # check that spatial extent is defined
        if spatialExtent is not None:
            # construct the spatial bounding box from inputs
            try:
                spatialExtent = [str(i).strip() for i in spatialExtent]
                bbox = ','.join(spatialExtent)
                payload['bounding_box'] = bbox

            except Exception as e:
                raise ValueError('Error parsing the spatialExtent information.\n '
                                 'See the followign error for details: {}'.format(e))

        # set the request parameters to class object
        self.payload = payload

        return

    def __repr__(self):
        strRepr = pformat(self.items, depth=2)
        return f'CMR Query with the following arguments:\n{strRepr}'

    @property
    def collections(self):
        """
        Property of a Query will return a list of OrderedDicts that
        contain the metadata information for all collections that meet
        the criteria of query

        Return:
            result.Collections: list of OrderedDicts with metadata information for collections
        """

        queryUrl = os.path.join(
            self.BASE_URL, "collections.{}".format(self.FORMAT))

        data = self._send_request(queryUrl)

        return Collections(data)

    @property
    def granules(self):
        """
        Property of a Query will return a list of OrderedDicts that
        contain the metadata information for all collections that meet
        the criteria of query

        Returns:
            result.Granules: list of OrderedDicts with metadata information for granules
        """

        queryUrl = os.path.join(
            self.BASE_URL, "granules.{}".format(self.FORMAT))

        data = self._send_request(queryUrl)

        return Granules(data)

    def _send_request(self, url):
        """
        Helper function that wraps a request and returns the results

        Args:
            url (string): URL string of the website to request data from

        Returns:
            data (dict): Dictionary of data returned from server

        Raises:
            requests.HTTPError: Raises error when request returns an error or has
                a status code that is not 200
        """

        try:
            r = requests.get(url, params=self.payload)
        except Exception as e:
            raise requests.HTTPError(e)

        if r.status_code == 200:
            data = r.json()
        else:
            raise requests.HTTPError('Provided query "{}" return error code {}'.format(
                r.url, r.status_code)
            )

        return data



import setuptools
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='simplecmr',
      version='0.0.1',
      description='Simple CMR',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='http://github.com/kmarkert/simple-cmr',
      packages=setuptools.find_packages(),
      author='Kel Markert',
      author_email='kel.markert@gmail.com',
      license='MIT',
      zip_safe=False,
      include_package_data=False,
      install_requires=[
          'requests',
          'requests-cache'
      ],
      #entry_points={
      #  'console_scripts': [
      #      'simplecmr = simplecmr.cli:main',
      #  ],
      #},
)



import fire

from simplecmr import query

class CLI(query.Query):
    def __init__(self,*args,**kwargs):
        super(CLI, self).__init__(*args,**kwargs)
        return

    def fetch



def main():
    fire.Fire(CLI)
